/* Preforking lib.
 * 
 * The source code may be redistributed, as long it remains intact, and
 * with this License included. You may modify or use this source code in
 * anyway you see fit, but credit must be given where applicable.
 *
 * This software comes with NO warranty, and the author is NOT liable for
 * any damages which may come from the use of this software.
 *
 * You may not use the name of the author to promote this software.
 * ---
 * 
 * I intentionaly dropped it into one huge file.
 * This way I keep everything static except init_daemon
 */

#define PREFORK_SOURCE

#include <config.h>
#include <prefork.h>
  
static unsigned int child_num = 0;
static unsigned int max = 0;
static unsigned int initial = 0;
static unsigned int cycle = 0;
static unsigned int busy = 0;

static int (*child_func)(int);
static char *logstring;

static int lock_fd;
static char buff[BUFFSIZE];
static int opened_lock = 0;

static struct child *children = NULL;
static int main_sockd;
static sigset_t mask;
static pid_t parent_pid;

struct sockaddr_in peer;

static pid_t wait_dead(void)
{
  int status;
  
  return(wait3(&status,WNOHANG,NULL));
}
static void wait_for_all_dead(void)
{
  int status;
  
  while(child_num--)  {
    wait(&status);
  }
  return;
}

static int fork_child(int num)
{
  int sv[2];
  pid_t  child_pid;
  int wanted = num;
  
  while(num--)  {
    if(socketpair(AF_UNIX,SOCK_STREAM,0,sv) < 0)  {
      error_log("socketpair: %s",strerror(errno));
      shutdown_server(0);
    }
    fcntl(sv[0],F_SETFL,O_NONBLOCK);
    fcntl(sv[1],F_SETFL,O_NONBLOCK);
    
    child_pid = fork();
    
    if(child_pid == -1)  {
#ifdef EAGAIN
      if(errno == EAGAIN) {
	close(sv[0]);
	close(sv[1]);
	return (wanted - num);
      } else {
	error_log("fatal fork: %s",strerror(errno));
	shutdown_server(0);
      }
#else
      error_log("fatal fork: %s",strerror(errno));
      shutdown_server(0);
#endif
    } else if(child_pid == 0) {
      struct sigaction childsig;
      
      sigemptyset(&childsig.sa_mask);
      childsig.sa_flags = 0;
      childsig.sa_handler = SIG_DFL;
      
      close(sv[0]);
      close_log();
      sigprocmask(SIG_SETMASK,&mask,NULL);
      sigaction(SIGTERM,&childsig,NULL);
#ifdef SIGCHLD
#define CHILD SIGCHLD
#elif SIGCLD
#define CHILD SIGCLD
#endif
      sigaction(CHILD,&childsig,NULL);
#undef CHILD
      child_main(sv[1]);
    } else {
      close(sv[1]);
      add_child_to_list(sv[0],child_pid);
      child_num++;
    }
  }
  return wanted;
}

static void child_main(int sockd)
{
  int length = sizeof(struct sockaddr_in);
  char message;
  int sockd2;
  
  init_log(logstring);
  open_lock();
  
  while(cycle--)  {
    message = READY;
    if(send_socket(sockd,(char *)&message,sizeof(message)))
      _exit(-1);
    kill(parent_pid,SIGUSR1);

    
    if(get_lock() < 0)  {
      error_log("Could'nt obtain lock: %s",strerror(errno));
      _exit(-1);
    }
    
    if((sockd2 = accept(main_sockd,(struct sockaddr *)&peer,&length)) < 0)  {
      error_log("accept: %s",strerror(errno));
      _exit(-1);
    }
    
    if(release_lock() < 0)  {
      error_log("Could'nt release lock: %s",strerror(errno));
      _exit(-1);
    }
    
    message = BUSY;
    if(send_socket(sockd,(char *)&message,sizeof(message)) < 0)
      _exit(-1);
    kill(parent_pid,SIGUSR1);

    child_func(sockd2);
    close(sockd2);
  }
  _exit(0);
}

void init_daemon(char *logstr,int num,int max_children,int num_cycle,int perfork,
	    int (*func)(int),int sockd)
{
  int retval;

  main_sockd = sockd;
  logstring = logstr;
  initial = num;
  max = max_children;
  cycle = num_cycle;
  child_func = func;
  parent_pid = getpid();
  
  init_log(logstring);

  mask = block_signals();
  setup_signal_handlers();

  init_lock();
  

  while(1) {

    if(child_num < initial) {
      retval = fork_child(initial - child_num);
    }
    
    if(busy == child_num) {
      if(max > busy)  {
	if((max - child_num) > perfork)
	  fork_child(perfork);
	else
	  fork_child((max - child_num));
      }
    }
    sigsuspend(&mask);
  }
  close_log();
}

static void child_died(int i)
{
  pid_t child_pid;
  
  while((child_pid = wait_dead()) > 0)  {
    child_num--;
    if(delete_child_from_list(child_pid) == BUSY)
      busy--;
    
  }
  return;
}

static void kill_all(void)
{
  struct child *ptr;
  
  for(ptr = children;ptr;ptr = ptr->next)  {
    kill(ptr->child_pid,SIGTERM);
  }
  wait_for_all_dead();
  return;
}

static void add_child_to_list(int sockd,pid_t child_pid)
{
  struct child *ptr;
  
  if(children == NULL)  {
    if((children = malloc(sizeof(struct child))) == NULL)  {
      error_log("malloc: %s",strerror(errno));
      shutdown_server(0);
    }
    ptr = children;
    ptr->next = NULL;
    
  } else {
    if((ptr = malloc(sizeof(struct child))) == NULL)  {
      error_log("malloc: %s",strerror(errno));
      shutdown_server(0);
    }
    ptr->next = children;
    children = ptr;
  }
  
  ptr->sockd = sockd;
  ptr->child_pid = child_pid;
  ptr->status = READY;
  return;
}

static unsigned char delete_child_from_list(pid_t child_pid)
{
  struct child *ptr, *last;
  char status;
  
  if(children->child_pid == child_pid)  {
    ptr = children->next;
    close(children->sockd);
    status = children->status;
    free(children);
    children = ptr;
  } else {
    for(ptr = children;ptr->next->child_pid != child_pid;ptr = ptr->next);
    last = ptr->next;
    ptr->next = ptr->next->next;
    close(last->sockd);
    status = last->status;
    free(last);
  }
  return status;
}

/* Attempt to create a unique file to use as a locking file. */

static void init_lock(void)
{
  int fd;
  FILE *fp;
  char *ptr;

  if((ptr = getenv("TMPDIR")) == NULL) {
    ptr = "/tmp";
  }

  strncpy(buff,ptr,(sizeof(buff)-1));
  if(buff[(strlen(buff)-1)] != '/') 
    strncat(buff,"/",(sizeof(buff)-1-strlen(buff)));

  strncat(buff,"daemon.XXXX",(sizeof(buff)-1-strlen(buff)));
  
  if((ptr = mktemp(buff)) == NULL)  {
    error_log("mktemp: %s",strerror(errno));
    exit(-1);
  }
  
  if((fd = open_safe(ptr)) < 0) {
    unlink(ptr);
    if((fd = open_safe(ptr)) < 0) {
      error_log("Couldnt open lock file: %s",strerror(errno));
      exit(-1);
    }
  }
  if((fp = fdopen(fd,"w")) == NULL) {
    error_log("fdopen: %s",strerror(errno));
    remove_lock_file();
    exit(-1);
  }

  if(fprintf(fp,"%u\n",getpid()) < 0)  {
    error_log("fprintf: %s",strerror(errno));
    remove_lock_file();
  }
   
  fclose(fp);
  close(fd);
  
  opened_lock = 1;
  return;
}

static void remove_lock_file(void)
{
  unlink(buff);
  return;
}

static void open_lock(void)
{
  int fd;
  
  if((fd = open(buff,O_RDWR)) < 0)  {
    error_log("Could'nt open lock file: %s",strerror(errno));
    exit(-1);
  }
  lock_fd = fd;
  return;
}
static void init_log(char *s)
{
  openlog(s,LOG_OPT,LOG_FACILITY);
  return;
}

static void close_log(void)
{
  closelog();
  return;
}

static void error_log(char *m,...)
{
  va_list ap;
  va_start(ap,m);

  vsyslog(LOG_LEVEL,m,ap);
  return;
}
/* Attempt to open a file safely. */

static int open_safe(char *path)
{
  return(open(path,O_RDWR|O_EXCL|O_CREAT,0600));
}

static sigset_t block_signals(void)
{
  sigset_t mask;
  sigset_t old;

  sigfillset(&mask);
  sigdelset(&mask,SIGKILL);
  sigdelset(&mask,SIGSTOP);
  sigdelset(&mask,SIGUSR1);
  
  sigprocmask(SIG_SETMASK,&mask,&old);
  
  return old;
}

static void setup_signal_handlers(void)
{
  struct sigaction mysig;
  
  sigfillset(&mysig.sa_mask);
  sigdelset(&mysig.sa_mask,SIGSTOP);
  sigdelset(&mysig.sa_mask,SIGKILL);

  mysig.sa_flags = 0;
  
  mysig.sa_handler = shutdown_server;
  
  if(sigaction(SIGTERM,&mysig,NULL) < 0)  {
    error_log("sigaction: %s",strerror(errno));
    shutdown_server(0);
  }
  
  mysig.sa_handler = check_children;
  
  if(sigaction(SIGUSR1,&mysig,NULL) < 0)  {
    error_log("sigaction: %s",strerror(errno));
    shutdown_server(0);
  }

#ifdef SA_NOCLDSTOP
  mysig.sa_flags = SA_NOCLDSTOP;
#endif
  mysig.sa_handler = child_died;
  
#ifdef SIGCHLD
#define CHILD SIGCHLD
#elif SIGCLD
#define CHILD SIGCLD
#endif
  if(sigaction(CHILD,&mysig,NULL) < 0)  {
    error_log("sigaction: %s",strerror(errno));
    shutdown_server(0);
  }
#undef CHILD
  
  return;
}

static void shutdown_server(int i)
{
  kill_all();
  if(opened_lock)
    remove_lock_file();
  exit(0);
}

static int send_socket(int sockd,char *mesg,int size)
{
  fd_set sd;
  int retval;

  FD_ZERO(&sd);
  FD_SET(sockd,&sd);
  
  retval = select((sockd+1),NULL,&sd,NULL,NULL);

  if(retval < 0)
    return -1;

retry:

  if(send(sockd,mesg,size,0x0) < 0) {
#ifdef ENOSR
    if(errno == ENOSR)
      goto retry;
#elif ENOMEM
    if(errno == ENOMEM)
      goto retry;
#endif
    if(errno == EAGAIN)
      goto retry;

    error_log("send: %s",strerror(errno));
    return -1;
  }

  return 0;
}

static int recv_socket(int sockd,char *buff,int size)
{
  int nbytes = 0;

  while(nbytes < size) {
    if(recv(sockd,(buff+nbytes),1,0x0) < 0) {
      if(errno == EAGAIN)
	return (nbytes-1);
      else
	return -1;
    }
    nbytes++;
  }
  return nbytes;
}

static void check_children(int i)
{
  fd_set rd;
  struct child *ptr;
  int retval, max = 0;
  struct timeval timeout = { 0, 0 };
  char message;
  
  FD_ZERO(&rd);
  for(ptr = children;ptr;ptr = ptr->next)  {
    FD_SET(ptr->sockd,&rd);
    if(max < ptr->sockd)  {
      max = (ptr->sockd +1);
    }
  }
  
  retval = select(max,&rd,NULL,NULL,&timeout);
  
  ptr = children;
  
  while(retval--)  {
    for(;ptr;ptr = ptr->next)  {
      if(FD_ISSET(ptr->sockd,&rd))  {
	FD_CLR(ptr->sockd,&rd);
	if(recv_socket(ptr->sockd,&message,sizeof(message)) < 0)  {
	  /* Bogus message the child could of died.
	   * Just close it and continue. */
	  close(ptr->sockd);
	  break;
	}

	
	if((message == READY) && (ptr->status == BUSY))  {
	  busy--;
	}
	else if((message == BUSY) && (ptr->status == READY))  {
	  busy++;
	}
	ptr->status = message;
	break;
      }
    }
  }
  return;
}

static int get_lock(void)
{
  return(flock(lock_fd,LOCK_EX));
}

static int release_lock(void)
{
  return(flock(lock_fd,LOCK_UN));
}
